/*
	LogicBoost Logic Analyzer for TI MSP430 Launchpad
	
    Copyright (C) 2012  Kevin Timmerman

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <msp430.h>
#include <stdint.h>

uint8_t cap_buf[384];									// Internal capture buffer

static const unsigned long smclk_freq = 16000000;		// SMCLK frequency in hertz
static const unsigned bps = 9600;						// Async serial bit rate

typedef struct {
	unsigned long read_count;
	unsigned long delay_count;
	unsigned flags;
	unsigned long divider;
	unsigned trig_mask;
	unsigned trig_value;
} config_t;
 
/*
P1.0	LED
P1.1	Serial Rx
P1.2	Serial Tx
P1.3	Switch
P1.4	RAM CLK
P1.5	RAM /CS
P1.6	LED
P1.7	Probe Latch Enable
P2.0 -> P2.5 RAM Data I/O, Probe data
P2.6 -> P2.7 32 kHz xtal
*/

// PORT1
static const uint8_t LED_RED = BIT0;
static const uint8_t SER_RX = BIT1;
static const uint8_t SER_TX = BIT2;
static const uint8_t PB_SW = BIT3;
static const uint8_t RAM_CLK = BIT4;
static const uint8_t RAM_CS  = BIT5;
static const uint8_t LED_GRN = BIT6;
static const uint8_t PROBE_EN = BIT7;

#define SC_RESET					0x00
#define SC_RUN						0x01
#define SC_ID						0x02
#define SC_XON						0x11
#define SC_XOFF						0x13
#define SC_SET_TRIGGER_MASK_0		0xC0
#define SC_SET_TRIGGER_MASK_1		0xC4
#define SC_SET_TRIGGER_MASK_2		0xC8
#define SC_SET_TRIGGER_MASK_3		0xCC
#define SC_SET_TRIGGER_VALUES_0		0xC1
#define SC_SET_TRIGGER_VALUES_1		0xC5
#define SC_SET_TRIGGER_VALUES_2		0xC9
#define SC_SET_TRIGGER_VALUES_3		0xCD
#define SC_SET_TRIGGER_CONFIG_0		0xC2
#define SC_SET_TRIGGER_CONFIG_1		0xC6
#define SC_SET_TRIGGER_CONFIG_2		0xCA
#define SC_SET_TRIGGER_CONFIG_3		0xCE
#define SC_SET_DIVIDER				0x80
#define SC_SET_READ_DELAY_COUNT		0x81
#define SC_SET_FLAGS				0x82

unsigned rxb(void)
{
	while(!(IFG2 & UCA0RXIFG));
	const unsigned n = UCA0RXBUF;
	IFG2 &= ~UCA0RXIFG;
	return n;
}

void txb(unsigned c)
{
	while(!(IFG2 & UCA0TXIFG));
	IFG2 &= ~UCA0TXIFG;
	UCA0TXBUF = c;
}

void txs(char *s) { while(*s) txb(*s++); }

void spi_write(unsigned long d, unsigned n)
{
	P1OUT &= ~RAM_CS;
	P2DIR = 0x3F;
	P2REN = 0;
	unsigned long m = 1L << (n - 1);
	while(m) {
		(d & m) ? (P2OUT = 0x3F) : (P2OUT = 0);
		m >>= 1;
		P1OUT |= RAM_CLK;
		if(m & 1) P2REN = 0x3F;
		P1OUT &= ~RAM_CLK;
	}
	P2DIR = 0;
	P2OUT = 0x3F;
}

void ram_clear(unsigned long n)
{
	spi_write(0x020000L, 24);
	P2DIR = 0x3F;
	P2REN = 0;
	P2OUT = 0;
	for( ; n; --n) {
		P1OUT |= RAM_CLK;
		P1OUT &= ~RAM_CLK;
	}
	P1OUT |= RAM_CS;
	P2DIR = 0;
	P2REN = 0x3F;
	P2OUT = 0x3F;
}

void setup_clock_norm(void)
{
	DCOCTL = 0;									// Switch to 16 MHz clock
	BCSCTL1 = CALBC1_16MHZ;						//
	DCOCTL  = CALDCO_16MHZ;						//
	BCSCTL2 = 0;								//
	__delay_cycles(150000);						// Wait for clock to settle
												/// todo: determine how long this delay should be
}

void setup_clock(unsigned c, unsigned d)		// Setup clock source, frequency and divider
{
	DCOCTL = 0;
	switch(c) {									// Clock source and frequency
		case 0:									// 0 == 16 MHz DCO
			BCSCTL1 = CALBC1_16MHZ;
			DCOCTL  = CALDCO_16MHZ;
			break;
		case 1:									// 1 == 1 MHz DCO
			BCSCTL1 = CALBC1_1MHZ;
			DCOCTL  = CALDCO_1MHZ;
			break;
												/// todo: 2 == 32.768 kHz xtal
	}
	BCSCTL2 = (d << 4) | (d << 1);				// Setup divider for MCLK and SMCLK
												// 0 == /1
												// 1 == /2
												// 2 == /4
												// 3 == /8
												
	__delay_cycles(50000);						// Wait for clock to settle												
}

void setup_clock_spi(unsigned long div)
{
	unsigned c, d;
	switch(div) {
		case 5:			c = 0; d = 0; break;	// 16 MHz
		case 11:		c = 0; d = 1; break;	// 8 MHz
		case 24:		c = 0; d = 2; break;	// 4 MHz
		case 49:		c = 0; d = 3; break;	// 2 MHz
		case 99:		c = 1; d = 0; break;	// 1 MHz
		case 199:		c = 1; d = 1; break;	// 500 kHz
		case 399:		c = 1; d = 2; break;	// 250 kHz
		case 799:		c = 1; d = 3; break;	// 125 kHz
		//case 3050:	c = 2; d = 0; break;	// 32.768 kHz
		//case 6102:	c = 2; d = 1; break;	// 16.384 kHz
		//case 12206:	c = 2; d = 2; break;	// 8.192 kHz
		//case 24413:	c = 2; d = 3; break;	// 4.096 kHz
												// - Hacks for original SUMP client
		case 4:			c = 0; d = 0; break;	//  20 MHz ->  16 MHz
		case 9:			c = 0; d = 1; break;	//  10 MHz ->   8 MHz
		case 19:		c = 0; d = 2; break;	//   5 MHz ->   4 MHz
		case 499:		c = 1; d = 2; break;	// 200 kHz -> 250 kHz
		case 999:		c = 1; d = 3; break;	// 100 kHz -> 125 kHz
		default:		return;
	}
	setup_clock(c, d);
}

void setup_clock_int(unsigned long div)
{
	unsigned c, d;
	switch(div) {
		case 49:		c = 0; d = 0; break;	// 2 MHz
		case 99:		c = 0; d = 1; break;	// 1 MHz
		case 199:		c = 0; d = 2; break;	// 500 kHz
		case 399:		c = 0; d = 3; break;	// 250 kHz
		case 799:		c = 1; d = 0; break;	// 125 kHz
		case 1599:		c = 1; d = 1; break;	// 62.5 kHz
		case 3199:		c = 1; d = 2; break;	// 31.25 kHz
		case 6399:		c = 1; d = 3; break;	// 15.625 kHz
		//case 24413:	c = 2; d = 0; break;	// 4.096 kHz
		//case 48827L:	c = 2; d = 1; break;	// 2.048 kHz
		//case 97655L:	c = 2; d = 2; break;	// 1.024 kHz
		//case 195311L:	c = 2; d = 3; break;	// 512 Hz
												// - Hacks for original SUMP client
		case 499:		c = 0; d = 3; break;	// 200 kHz -> 250 kHz
		case 999:		c = 1; d = 0; break;	// 100 kHz -> 125 kHz
		case 1999:		c = 1; d = 1; break;	//  50 kHz ->  62.5 kHz
		case 4999:		c = 1; d = 2; break;	//  20 kHz ->  31.25 kHz
		case 9999:		c = 1; d = 3; break;	//  10 kHz ->  15.625 kHz	
		default:		return;
	}
	setup_clock(c, d);
}

void send2host_test(unsigned long count)
{
	while(count--) txb(count);
}

void send2host(unsigned addr, unsigned long count)
{
	spi_write(0x030000L | addr, 24); 			// Read starting at addr
	P2DIR = 0;									// Switch RAM bus to input
	P2REN = 0x3F;								// Pull up all data lines
	P2OUT = 0x3F;								//
	unsigned long n = count;					//
	for( ; n; --n) {							// Do requested length
		txb(P2IN);								// Send sample to host
		P1OUT |= RAM_CLK;						// Clock RAM
		P1OUT &= ~RAM_CLK;						//
	}											//
	P1OUT |= RAM_CS;							// Disable RAM
}

void capture_spi(config_t * const config)
{
	setup_clock_spi(config->divider);
	
	const unsigned pre_trig = ((config->read_count - config->delay_count) + 32 + 7) >> 3;
	const unsigned post_trig = (config->delay_count + 32 + 7) >> 3;	

	spi_write(0x020000L, 24);					// Prepare SPI RAM for write beginning at address 0
	P1OUT &= ~PROBE_EN;							// Enable probe buffer
	P1SEL |= RAM_CLK;							// Begin clocking SPI RAM
	TA1CTL |= TACLR;							// Reset timer to sync to SPI RAM address
												//
	while(TA1R < pre_trig);						// Wait for SPI RAM to fill with pre-trigger data
	if(config->trig_mask) {						// Is there a trigger mask?
		P2IES = ~config->trig_value & 0x3F;		// Setup pin change bitmask
		P2IFG = 0;								// Clear pin change flags
		while(!(P2IFG & config->trig_mask));	// Wait for pin change on pins of interest
	}											//
	const unsigned start = TA1R;				// Get address of trigger
	TA1CCR0 = start + post_trig;				// Setup capture end address
	TA1CCTL0 = 0;								// Reset timer interrupt flag
	while(!(TA1CCTL0 & CCIFG));					// Wait for timer interrupt
												//
	P1SEL &= ~RAM_CLK;							// Stop SPI RAM clock
	P1OUT |= RAM_CS;							// Disable SPI RAM
	P1OUT |= PROBE_EN;							// Disable probe buffer
												//
	setup_clock_norm();							// Switch to clock for UART
												// Send capture to host
	send2host(start - pre_trig, config->read_count);
}

#pragma vector = PORT2_VECTOR
__interrupt void port2vector(void)
{
												// - Get 2 samples to buffer for each loop iteration
												//   Exactly 8 clock cycles per sample
												//   So max sample rate with 16 MHz clock is 2 Msps
	__asm("loop:");								//
	__asm(" mov.b &P2IN, 0(R15)");				// 4 Get sample to buffer
	__asm(" incd R15");							// 1 Increment sample pointer by one word
	__asm(" dec R14");							// 1 Decrement sample count
	__asm(" mov.b &P2IN, -1(R15)");				// 4 Get sample to buffer
	__asm(" jne loop");							// 2 Loop until all samples are done
	__asm(" clr.b &P2IE");						// Disable all Port2 interupts
   	__asm(" bic #0x18, 0(R1)");					// Turn off global interrupts and wakeup main code
   
}

void capture_internal(config_t * const config)
{
	setup_clock_int(config->divider); 			// Setup clock
												//
	P1OUT |= RAM_CS;							// Make sure SPI RAM is disabled
	P1OUT &= ~PROBE_EN;							// Enable probe buffer
	P1OUT |= RAM_CLK;							// Set probe latch to passthru
												//
	if(config->trig_mask) {						//											
		P2IES = ~config->trig_value & 0x3F;		// Setup pin change direction bitmask
		P2IFG = 0;								// Clear pin change flags
		P2IE  = config->trig_mask & 0x3F;		// Setup pin change interrupt mask
		__asm(" push R14");						// Save registers used by ISR
		__asm(" push R15");						//
		__asm(" mov  #384 / 2, R14");			// Setup sample count - 2 samples per loop iteration
		__asm(" mov  #cap_buf, R15");			// Setup buffer pointer
		__asm(" bis  #0x18, R2");				// Enable interrupts and sleep in LPM0 until ISR has completed
		__asm(" pop  R15");						// Restore registers
		__asm(" pop  R14");						//
	} else {									//
		__asm(" push R14");						// Save registers used by ISR
		__asm(" push R15");						//
		__asm(" mov  #384 / 2, R14");			// Setup sample count - 2 samples per loop iteration
		__asm(" mov  #cap_buf, R15");			// Setup buffer pointer
												// - Call ISR from mainline context
		__asm(" push #return");					// Push return address
		__asm(" push R2");						// Push status register
		__asm(" br   #port2vector");			// Jump to ISR - it will pop SR and PC
		__asm("return:");						// Return to here
		__asm(" pop  R15");						// Restore registers
		__asm(" pop  R14");						//
	}											//
												//
	P1OUT &= ~RAM_CLK;							// Turn off probe latch passthru
	P1OUT |= PROBE_EN;							// Disable probe buffer
												//
	setup_clock_norm();							// Switch back to normal clock												
												//
	unsigned count = config->read_count;		// Get sample count
	if(count == 384) {							// - OLS client with LogicBoost configuration
		if(config->trig_mask) {					// If trigger was used, then send a fake sample
			txb(~config->trig_value & 0x3F);	//  to make protocol analyzer happy
			 --count;							//
		}										//
		uint8_t *b = cap_buf;					// Setup pointer to capture buffer
		for( ; count; --count) txb(*b++);		// Send samples to host
	} else if(count == 512) {					// - Original SUMP client
												// Get group flags
		const unsigned groups = ((config->flags >> 2) & 0x0F) ^ 0x0F;
		unsigned n = 128;						// Padding to make 512 total samples
		if(config->trig_mask) --n;				// One less pad sample if trigger was used
		for( ; n; --n) {						// Send padding
			if(groups & 1) txb(0);				//
			if(groups & 2) txb(0);				//
			if(groups & 4) txb(0);				//
			if(groups & 8) txb(0);				//
		}										//
		n = 384;
		do {									// Send samples in reverse order
			--n;								//
			if(groups & 1) txb(cap_buf[n]);		//
			if(groups & 2) txb(0);				//
			if(groups & 4) txb(0);				//
			if(groups & 8) txb(0);				//
		} while(n);								//
		if(config->trig_mask) {					// If trigger was used, then send a fake sample
												//  to make protocol analyzer happy
			if(groups & 1) txb(~config->trig_value & 0x3F);
			if(groups & 2) txb(0);				//
			if(groups & 4) txb(0);				//
			if(groups & 8) txb(0);				//
		}										//
	}
}

// - SUMP protocol documentation
// http://www.sump.org/projects/analyzer/protocol/
// http://dangerousprototypes.com/docs/The_Logic_Sniffer%27s_extended_SUMP_protocol
//
void sump(void)									// Handle SUMP logic analyzer protocol
{
	unsigned cmd, lc[4], lclw;
	config_t config = { 0 };
	
	for(;;) {
		cmd = rxb();
		switch(cmd) {
			case SC_RESET:
				//spi_write(0x0140, 16);
				//P1OUT |= RAM_CS;
				break;
				
			case SC_RUN:
				switch(config.read_count) {
					case 384:						// OLS client - Internal memory
					case 512:						// SUMP client - Internal memory
						capture_internal(&config);
						break;
					default:
						capture_spi(&config);		// OLS client - SPI memory
						break;
				}
				break;
				
			case SC_ID:
				txs("1ALS");
				break;
				
			case 'i':									// For checking comms with terminal
				txs("MSP430 Logic Analyzer\r\n");
				break;
				
			//case SC_XON:  break;
			//case SC_XOFF: break;
			
			case SC_SET_TRIGGER_MASK_0:
			//case SC_SET_TRIGGER_MASK_1:
			//case SC_SET_TRIGGER_MASK_2:
			//case SC_SET_TRIGGER_MASK_3:
			case SC_SET_TRIGGER_VALUES_0:
			//case SC_SET_TRIGGER_VALUES_1:
			//case SC_SET_TRIGGER_VALUES_2:
			//case SC_SET_TRIGGER_VALUES_3:
			case SC_SET_TRIGGER_CONFIG_0:
			//case SC_SET_TRIGGER_CONFIG_1:
			//case SC_SET_TRIGGER_CONFIG_2:
			//case SC_SET_TRIGGER_CONFIG_3:
			case SC_SET_DIVIDER:
			case SC_SET_READ_DELAY_COUNT:
			case SC_SET_FLAGS:
				lc[0] = rxb();
				lc[1] = rxb();
				lclw = lc[0] | (lc[1] << 8);
				lc[2] = rxb();
				lc[3] = rxb();
				switch(cmd) {
					case SC_SET_TRIGGER_MASK_0:
						config.trig_mask = lclw;
						break;
					//case SC_SET_TRIGGER_MASK_1:	break;
					//case SC_SET_TRIGGER_MASK_2:	break;
					//case SC_SET_TRIGGER_MASK_3:	break;
					
					case SC_SET_TRIGGER_VALUES_0:
						config.trig_value = lclw;
						break;
					//case SC_SET_TRIGGER_VALUES_1:	break;
					//case SC_SET_TRIGGER_VALUES_2:	break;
					//case SC_SET_TRIGGER_VALUES_3:	break;
					
					case SC_SET_TRIGGER_CONFIG_0:	break;
					//case SC_SET_TRIGGER_CONFIG_1:	break;
					//case SC_SET_TRIGGER_CONFIG_2:	break;
					//case SC_SET_TRIGGER_CONFIG_3:	break;
					
					case SC_SET_DIVIDER:
						config.divider = lc[2];
						config.divider <<= 16;
						config.divider |= lclw;
						break;
						
					case SC_SET_READ_DELAY_COUNT:
						config.read_count = lclw;
						++config.read_count;
						config.read_count <<= 2;
						config.delay_count = lc[2] | (lc[3] << 8);
						config.delay_count <<= 2;
						break;
						
					case SC_SET_FLAGS:
						config.flags = lc[0];
						break;
				}
				break;
			default:
				break;
		}
		P1OUT ^= LED_RED;
	}
}

void main(void)
{
	const unsigned long brd = (smclk_freq + (bps >> 1)) / bps; // Bit rate divisor
	
	WDTCTL = WDTPW + WDTHOLD; 					// No watchdog reset
												//
	setup_clock_norm();							// Setup clock												
												//
												// Setup IO
	P1OUT = SER_RX | SER_TX | PB_SW | RAM_CS | PROBE_EN;
	P1DIR = LED_RED | SER_TX | RAM_CLK | RAM_CS | LED_GRN | PROBE_EN;
	P1REN = SER_RX | PB_SW;						//
	P1SEL  = BIT1 | BIT2;						//
	P1SEL2 = BIT1 | BIT2;						//
												//
	P2OUT = 0x3F;								//
	P2DIR = 0x00;								//
	P2REN = 0x3F;								//
												//
												//
	UCA0CTL1 = UCSWRST;							// Hold USCI in reset to allow configuration
	UCA0CTL0 = 0;								// No parity, LSB first, 8 bits, one stop bit, UART (async)
	UCA0BR1 = (brd >> 12) & 0xFF;				// High byte of whole divisor
	UCA0BR0 = (brd >> 4) & 0xFF;				// Low byte of whole divisor
	UCA0MCTL = ((brd << 4) & 0xF0) | UCOS16;	// Fractional divisor, oversampling mode				
	UCA0CTL1 = UCSSEL_2;						// Use SMCLK for bit rate generator, release reset												

	TA1CTL = TASSEL_2 | ID_3 | MC_2;			// SMCLK, /8, Continuous
	
#if 1											// Test signal on P1.6
	P1SEL |= LED_GRN;
	TA0CCR0 = 48 - 1;
	TA0CCR1 = 16;
	TA0CCTL1 = OUTMOD_7;
	TA0CTL = TASSEL_2 | MC_1 | TACLR;	
#endif

	spi_write(0x0140, 16);
	P1OUT |= RAM_CS;

	sump();
}
